# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import random
from collections.abc import Mapping
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Union

import gym
import numpy as np
import torch
from gym.spaces import Box
from torch import nn

from ..core import Solution, SolutionBatch
from ..tools import Device, ReadOnlyTensor, pass_info_if_needed, split_workload
from .baseneproblem import BaseNEProblem
from .net import str_to_net
from .net.rl import ActClipWrapperModule, ObsNormWrapperModule
from .net.runningnorm import RunningNorm
from .net.vecrl import Policy, TorchWrapper, make_vector_env


def _is_numpy_array_entirely_inf(x: np.ndarray) -> bool:
    return np.all(np.isinf(x))


def _is_numpy_array_entirely_positive_inf(x: np.ndarray) -> bool:
    return _is_numpy_array_entirely_inf(x) and np.all(x > 0)


def _is_numpy_array_entirely_negative_inf(x: np.ndarray) -> bool:
    return _is_numpy_array_entirely_inf(x) and np.all(x < 0)


def _numpy_arrays_specify_no_bounds(low: np.ndarray, high: np.ndarray) -> bool:
    return _is_numpy_array_entirely_negative_inf(low) and _is_numpy_array_entirely_positive_inf(high)


def _numpy_arrays_specify_bounds(low: np.ndarray, high: np.ndarray) -> bool:
    return not _numpy_arrays_specify_no_bounds(low, high)


def _make_dummy_env(env: Union[str, Callable], **kwargs) -> gym.vector.VectorEnv:
    if isinstance(env, str):
        result = make_vector_env(env, num_envs=1, **kwargs)
    else:
        result = env(num_envs=1, **kwargs)
    result = TorchWrapper(result, discrete_to_continuous_act=True)
    return result


def _env_spaces(env: Union[str, Callable], **kwargs) -> tuple:
    tmp_env = _make_dummy_env(env, **kwargs)
    return tmp_env.single_observation_space, tmp_env.single_action_space


def _env_constants_for_str_net(env: Union[str, Callable], **kwargs) -> dict:
    obs_space, act_space = _env_spaces(env, **kwargs)
    return {
        "obs_length": obs_space.shape[0],
        "act_length": act_space.shape[0],
        "obs_shape": obs_space.shape,
        "obs_space": obs_space.shape,
        "act_shape": act_space.shape,
        "act_space": act_space.shape,
    }


def _env_constants_for_callable_net(env: Union[str, Callable], **kwargs) -> dict:
    obs_space, act_space = _env_spaces(env, **kwargs)
    return {
        "obs_length": obs_space.shape[0],
        "act_length": act_space.shape[0],
        "obs_shape": obs_space.shape,
        "obs_space": obs_space,
        "act_shape": act_space.shape,
        "act_space": act_space,
    }


class VecGymNE(BaseNEProblem):
    """
    An EvoTorch problem for solving vectorized gym environments
    """

    def __init__(
        self,
        env: Union[str, Callable],
        network: Union[str, Callable, nn.Module],
        *,
        env_config: Optional[Mapping] = None,
        max_num_envs: Optional[int] = None,
        network_args: Optional[Mapping] = None,
        observation_normalization: bool = False,
        decrease_rewards_by: Optional[float] = None,
        alive_bonus_schedule: Optional[tuple] = None,
        action_noise_stdev: Optional[float] = None,
        num_episodes: int = 1,
        device: Optional[Device] = None,
        num_actors: Optional[Union[int, str]] = "num_devices",
        num_gpus_per_actor: Optional[int] = None,
        num_subbatches: Optional[int] = None,
        subbatch_size: Optional[int] = None,
        actor_config: Optional[Mapping] = None,
    ):
        """
        Initialize the VecGymNE.

        Args:
            env: Environment to be solved.
                If this is given as a string starting with "gym::" (e.g.
                "gym::Humanoid-v4", etc.), then it is assumed that the target
                environment is a classical gym environment.
                If this is given as a string starting with "brax::" (e.g.
                "brax::humanoid", etc.), then it is assumed that the target
                environment is a brax environment.
                If this is given as a string which does not contain "::" at
                all (e.g. "Humanoid-v4", etc.), then it is assumed that the
                target environment is a classical gym environment. Therefore,
                "gym::Humanoid-v4" and "Humanoid-v4" are equivalent.
                If this argument is given as a Callable (maybe a function or a
                class), then, with the assumption that this Callable expects
                a keyword argument `num_envs: int`, this Callable is called
                and its result (expected as a `gym.vector.VectorEnv` instance)
                is used as the environment.
            network: A network structure string, or a Callable (which can be
                a class inheriting from `torch.nn.Module`, or a function
                which returns a `torch.nn.Module` instance), or an instance
                of `torch.nn.Module`.
                The object provided here determines the structure of the
                neural network whose parameters will be evolved.
                A network structure string is a string which can be processed
                by `evotorch.neuroevolution.net.str_to_net(...)`.
                Please see the documentation of the function
                `evotorch.neuroevolution.net.str_to_net(...)` to see how such
                a neural network structure string looks like.
                Note that this network can be a recurrent network.
                When the network's `forward(...)` method can optionally accept
                an additional positional argument for the hidden state of the
                network and returns an additional value for its next state,
                then the policy is treated as a recurrent one.
                When the network is given as a callable object (e.g.
                a subclass of `nn.Module` or a function) and this callable
                object is decorated via `evotorch.decorators.pass_info`,
                the following keyword arguments will be passed:
                (i) `obs_length` (the length of the observation vector),
                (ii) `act_length` (the length of the action vector),
                (iii) `obs_shape` (the shape tuple of the observation space),
                (iv) `act_shape` (the shape tuple of the action space),
                (v) `obs_space` (the Box object specifying the observation
                space, and
                (vi) `act_space` (the Box object specifying the action
                space). Note that `act_space` will always be given as a
                `gym.spaces.Box` instance, even when the actual gym
                environment has a discrete action space. This because
                `VecGymNE` always expects the neural network to return
                a tensor of floating-point numbers.
            env_config: Keyword arguments to pass to the environment while
                it is being created.
            max_num_envs: Maximum number of environments to be instantiated.
                By default, this is None, which means that the number of
                environments can go up to the population size (or up to the
                number of solutions that a remote actor receives, if the
                problem object is configured to have parallelization).
                For situations where the current reinforcement learning task
                requires large amount of resources (e.g. memory), allocating
                environments as much as the number of solutions might not
                be feasible. In such cases, one can set `max_num_envs` as an
                integer to bring an upper bound (in total, across all the
                remote actors, for when the problem is parallelized) to how
                many environments can be allocated.
            network_args: Any additional keyword argument to be used when
                instantiating the network can be specified via `network_args`
                as a dictionary. If there are no such additional keyword
                arguments, then `network_args` can be left as None.
                Note that the argument `network_args` is expected to be None
                when the network is specified as a `torch.nn.Module` instance.
            observation_normalization: Whether or not online normalization
                will be done on the encountered observations.
            decrease_rewards_by: If given as a float, each reward will be
                decreased by this amount. For example, if the environment's
                reward function has a constant "alive bonus" (i.e. a bonus
                that is constantly added onto the reward as long as the
                agent is alive), and if you wish to negate this bonus,
                you can set `decrease_rewards_by` to this bonus amount,
                and the bonus will be nullified.
                If you do not wish to affect the rewards in this manner,
                keep this as None.
            alive_bonus_schedule: Use this to add a customized amount of
                alive bonus.
                If left as None (which is the default), additional alive
                bonus will not be added.
                If given as a tuple `(t, b)`, an alive bonus `b` will be
                added onto all the rewards beyond the timestep `t`.
                If given as a tuple `(t0, t1, b)`, a partial (linearly
                increasing towards `b`) alive bonus will be added onto
                all the rewards between the timesteps `t0` and `t1`,
                and a full alive bonus (which equals to `b`) will be added
                onto all the rewards beyond the timestep `t1`.
            action_noise_stdev: If given as a real number `s`, then, for
                each generated action, Gaussian noise with standard
                deviation `s` will be sampled, and then this sampled noise
                will be added onto the action.
                If action noise is not desired, then this argument can be
                left as None.
                For sampling the noise, the global random number generator
                of PyTorch on the simulator's device will be used.
            num_episodes: Number of episodes over which each policy will
                be evaluated. The default is 1.
            device: The device in which the population will be kept.
                If you wish to do a single-GPU evolution, we recommend
                to set this as "cuda" (or "cuda:0", or "cuda:1", etc.),
                assuming that the simulator will also instantiate itself
                on that same device.
                Alternatively, if you wish to do a multi-GPU evolution,
                we recommend to leave this as None or set this as "cpu",
                so that the main population will be kept on the cpu
                and the remote actors will perform their evaluations on
                the GPUs that are assigned to them.
            num_actors: Number of actors.
                This is, by default, "num_devices" which creates multiple
                actors in such a way that each actor will be assigned
                an entire GPU.
                This default setting considers the multi-GPU scenario
                (and requires multiple GPUs to work correctly).
                If you wish to perform the entire evolution on a single
                GPU, we recommend to set `num_actors` as 0 (so that there
                will be no remote actors) and set `device` as "cuda"
                (or "cuda:0", or "cuda:1", etc., the device in which the
                simulator will be instantiated).
                If you do not have any GPUs the special value "num_devices"
                will allocate n actors, n being the number of CPUs.
                You can also set `num_actors` as an integer to determine
                the exact number of subprocesses which will evaluate the
                solutions in parallel. Note that, however, when you set
                `num_actors` as an integer, you will also have to explicitly
                specify the number of GPUs each actor will be assigned
                via the argument `num_gpus_per_actor`.
            num_gpus_per_actor: Number of GPUs to be assigned to each
                actor. This can be an integer or a float (for when you
                wish to assign fractional amounts of GPUs to actors).
                When `num_actors` has the special value "num_devices",
                the argument `num_gpus_per_actor` is expected to be left as
                None.
            num_subbatches: For when there are multiple actors, you can
                set this to an integer n if you wish the population
                to be divided exactly into n sub-batches. The actors, as they
                finish their currently assigned sub-batch of solutions,
                will pick the next un-evaluated sub-batch.
                If you specify too large numbers for this argument, then
                each sub-batch will be smaller.
                When working with vectorized simulators on GPU, having too
                many and too small sub-batches can hurt the performance.
                This argument can be left as None, in which case, assuming
                that `subbatch_size` is also None, the population will be
                split to m sub-batches, m being the number of actors.
            subbatch_size: For when there are multiple actors, you can
                set this to an integer n if you wish the population to be
                divided into sub-batches in such a way that each sub-batch
                will consist of exactly n solutions. The actors, as they
                finish their currently assigned sub-batch of solutions,
                will pick the next un-evaluated sub-batch.
                If you specify too small numbers for this argument, then
                there will be many sub-batches, each sub-batch having a
                small number of solutions.
                When working with vectorized simulators on GPU, having too
                many and too small sub-batches can hurt the performance.
                This argument can be left as None, in which case, assuming
                that `num_subbatches` is also None, the population will be
                split to m sub-batches, m being the number of actors.
            actor_config: Additional configuration to be used when creating
                each actor with the help of `ray` library.
                Can be left as None if additional configuration is not needed.
        """

        # Store the string or the Callable that will be used to generate the reinforcement learning environment.
        self._env_maker = env

        # Declare the variable which will store the environment.
        self._env: Optional[TorchWrapper] = None

        # Declare the variable which will store the batch size of the vectorized environment.
        self._num_envs: Optional[int] = None

        # Store the upper bound (if any) regarding how many environments can exist at the same time.
        self._max_num_envs: Optional[int] = None if max_num_envs is None else int(max_num_envs)

        # Actor-specific upper bound regarding how many environments can exist at the same time.
        # This variable will be filled by the `_parallelize(...)` method.
        self._actor_max_num_envs: Optional[int] = None

        # Declare the variable which stores whether or not we properly initialized the `_actor_max_num_envs` variable.
        self._actor_max_num_envs_ready: bool = False

        # Store the additional configurations to be used as keyword arguments while instantiating the environment.
        self._env_config: dict = {} if env_config is None else dict(env_config)

        # Declare the variable that will store the device of the simulator.
        # This variable will be filled when the first observation is received from the environment.
        # The device of the observation array received from the environment will determine the value of this variable.
        self._simulator_device: Optional[torch.device] = None

        # Store the neural network architecture (that might be a string or an `nn.Module` instance).
        self._architecture = network

        if network_args is None:
            # If `network_args` is given as None, change it to an empty dictionary
            network_args = {}

        if isinstance(network, str):
            # If the network is given as a string, then we will need the values for the constants `obs_length`,
            # `act_length`, and `obs_space`. To obtain those values, we use our helper function
            # `_env_constants_for_str_net(...)` which temporarily instantiates the specified environment and returns
            # its needed constants.
            env_constants = _env_constants_for_str_net(self._env_maker, **(self._env_config))
        elif isinstance(network, nn.Module):
            # If the network is an already instantiated nn.Module, then we do not prepare any pre-defined constants.
            env_constants = {}
        else:
            # If the network is given as a Callable, then we will need the values for the constants `obs_length`,
            # `act_length`, and `obs_space`. To obtain those values, we use our helper function
            # `_env_constants_for_callable_net(...)` which temporarily instantiates the specified environment and
            # returns its needed constants.
            env_constants = _env_constants_for_callable_net(self._env_maker, **(self._env_config))

        # Build a `Policy` instance according to the given architecture, and store it.
        if isinstance(network, str):
            instantiated_net = str_to_net(network, **{**env_constants, **network_args})
        elif isinstance(network, nn.Module):
            instantiated_net = network
        else:
            instantiated_net = pass_info_if_needed(network, env_constants)(**network_args)
        self._policy = Policy(instantiated_net)

        # Store the boolean which indicates whether or not there will be observation normalization.
        self._observation_normalization = bool(observation_normalization)

        # Declare the variables that will store the observation-related stats if observation normalization is enabled.
        self._obs_stats: Optional[RunningNorm] = None
        self._collected_stats: Optional[RunningNorm] = None

        # Store the number of episodes configuration given by the user.
        self._num_episodes = int(num_episodes)

        # Store the `decrease_rewards_by` configuration given by the user.
        self._decrease_rewards_by = None if decrease_rewards_by is None else float(decrease_rewards_by)

        if alive_bonus_schedule is None:
            # If `alive_bonus_schedule` argument is None, then we store it as None as well.
            self._alive_bonus_schedule = None
        else:
            # This is the case where the user has specified an `alive_bonus_schedule`.
            alive_bonus_schedule = list(alive_bonus_schedule)
            alive_bonus_schedule_length = len(alive_bonus_schedule)
            if alive_bonus_schedule_length == 2:
                # If `alive_bonus_schedule` was given as a 2-element sequence (t, b), then store it as (t, t, b).
                # This means that the partial alive bonus time window starts and ends at t, therefore, there will
                # be no alive bonus until t, and beginning with t, there will be full alive bonus.
                self._alive_bonus_schedule = [
                    int(alive_bonus_schedule[0]),
                    int(alive_bonus_schedule[0]),
                    float(alive_bonus_schedule[1]),
                ]
            elif alive_bonus_schedule_length == 3:
                # If `alive_bonus_schedule` was given as a 3-element sequence (t0, t1, b), then store those 3
                # elements.
                self._alive_bonus_schedule = [
                    int(alive_bonus_schedule[0]),
                    int(alive_bonus_schedule[1]),
                    float(alive_bonus_schedule[2]),
                ]
            else:
                # `alive_bonus_schedule` sequences with unrecognized lengths trigger an error.
                raise ValueError(
                    f"Received invalid number elements as the alive bonus schedule."
                    f" Expected 2 or 3 items, but got these: {self._alive_bonus_schedule}"
                    f" (having a length of {len(self._alive_bonus_schedule)})."
                )

        # If `action_noise_stdev` is specified, store it.
        self._action_noise_stdev = None if action_noise_stdev is None else float(action_noise_stdev)

        # Initialize the counters for the number of simulator interactions and the number of episodes.
        self._interaction_count: int = 0
        self._episode_count: int = 0

        # Call the superclass
        super().__init__(
            objective_sense="max",
            initial_bounds=(-0.00001, 0.00001),
            solution_length=self._policy.parameter_length,
            device=device,
            dtype=torch.float32,
            num_actors=num_actors,
            num_gpus_per_actor=num_gpus_per_actor,
            actor_config=actor_config,
            num_subbatches=num_subbatches,
            subbatch_size=subbatch_size,
        )

    def _parallelize(self):
        super()._parallelize()
        if self.is_main:
            if not self._actor_max_num_envs_ready:
                if self._actors is None:
                    self._actor_max_num_envs = self._max_num_envs
                else:
                    if self._max_num_envs is not None:
                        max_num_envs_per_actor = split_workload(self._max_num_envs, len(self._actors))
                        for i_actor, actor in enumerate(self._actors):
                            actor.call.remote("_set_actor_max_num_envs", max_num_envs_per_actor[i_actor])
                self._actor_max_num_envs_ready = True

    def _set_actor_max_num_envs(self, n: int):
        self._actor_max_num_envs = n
        self._actor_max_num_envs_ready = True

    @property
    def observation_normalization(self) -> bool:
        return self._observation_normalization

    def set_episode_count(self, n: int):
        """
        Set the episode count manually.
        """
        self._episode_count = int(n)

    def set_interaction_count(self, n: int):
        """
        Set the interaction count manually.
        """
        self._interaction_count = int(n)

    @property
    def interaction_count(self) -> int:
        """
        Get the total number of simulator interactions made.
        """
        return self._interaction_count

    @property
    def episode_count(self) -> int:
        """
        Get the total number of episodes completed.
        """
        return self._episode_count

    def _get_local_episode_count(self) -> int:
        return self.episode_count

    def _get_local_interaction_count(self) -> int:
        return self.interaction_count

    def _get_env(self, num_policies: int) -> TorchWrapper:
        # Get the existing environment instance stored by this VecGymNE, after (re)building it if needed.

        if (self._env is None) or (num_policies > self._num_envs):
            # If this VecGymNE does not have its environment ready yet (i.e. the `_env` attribute is None)
            # or it the batch size of the previously instantiated environment is not enough to deal with
            # the number of policies (i.e. the `_num_envs` attribute is less than `num_policies`), then
            # we (re)build the environment.

            # Keyword arguments to pass to the TorchWrapper.
            torch_wrapper_cfg = dict(
                force_classic_api=True,
                discrete_to_continuous_act=True,
                clip_actions=True,
            )

            if isinstance(self._env_maker, str):
                # If the environment is specified via a string, then we use our `make_vector_env` function.
                self._env = make_vector_env(
                    self._env_maker, num_envs=num_policies, **torch_wrapper_cfg, **(self._env_config)
                )
            else:
                # If the environment is specified via a Callable, then we call it.
                # We expect this Callable to accept a keyword argument named `num_envs`, and additionally, we pass
                # the environment configuration dictionary as keyword arguments.
                self._env = self._env_maker(num_envs=num_policies, **(self._env_config))

                if not isinstance(self._env, gym.vector.VectorEnv):
                    # If what is returned by the Callable is not a vectorized environment, then we trigger an error.
                    raise TypeError("This is not a vectorized environment")

                # We wrap the returned vectorized environment with a TorchWrapper, so that the actions that we send
                # and the observations and rewards that we receive are PyTorch tensors.
                self._env = TorchWrapper(self._env, **torch_wrapper_cfg)

            if self._env.num_envs != num_policies:
                # If the finally obtained vectorized environment has a different number of batch size, then we trigger
                # an error.
                raise ValueError("Incompatible number of environments")

            # We update the batch size of the created environment.
            self._num_envs = num_policies

            if not isinstance(self._env.single_observation_space, Box):
                # If the observation space is not Box, then we trigger an error.
                raise TypeError(
                    f"Unsupported observation type: {self._env.single_observation_space}."
                    f" Only Box-typed observation spaces are supported."
                )

            try:
                # If possible, use the `seed(...)` method to explicitly randomize the environment.
                # Although the new gym API removed the seed method, some environments define their own `seed(...)`
                # method for randomization.
                new_seed = random.randint(0, (2**32) - 1)
                self._env.seed(new_seed)
            except Exception:
                # Our attempt at manually seeding the environment has failed.
                # This could be because the environment does not have a `seed(...)` method.
                # Nothing to do.
                pass

        return self._env

    @property
    def _nonserialized_attribs(self):
        # Call the `_nonserialized_attribs` property implementation of the superclass to receive the base list
        # of non-serialized attributes, then add "_env" to this base list, and then return the resulting list.
        return super()._nonserialized_attribs + ["_env"]

    @property
    def _grad_device(self) -> torch.device:
        # For distributed mode, this property determines the device in which the temporary populations will be made
        # for gradient computation.

        if self._simulator_device is None:
            # If the simulator device is not known yet, then we return the cpu device.
            return torch.device("cpu")
        else:
            # If the simulator device is known, then we return that device.
            return self._simulator_device

    def _make_running_norm(self, observation: torch.Tensor) -> RunningNorm:
        # Make a new RunningNorm instance according to the observation tensor.
        # The dtype and the device of the new RunningNorm is taken from the observation.
        # This new RunningNorm is empty (i.e. does not contain any stats yet).
        return RunningNorm(shape=observation.shape[1:], dtype=observation.dtype, device=observation.device)

    def _transfer_running_norm(self, rn: RunningNorm, observation: torch.Tensor) -> RunningNorm:
        # Transfer (if necessary) the RunningNorm to the device of the observation tensor.
        # The returned RunningNorm may be the RunningNorm itself (if the device did not change)
        # or a new copy (if the device did change).
        if torch.device(rn.device) != torch.device(observation.device):
            rn = rn.to(observation.device)
        return rn

    def _normalize_observation(
        self, observation: torch.Tensor, *, mask: Optional[torch.Tensor] = None, update_stats: bool = True
    ) -> torch.Tensor:
        # This function normalizes the received observation batch.
        # If a mask is given (as a tensor of booleans), only observations with corresponding mask value set as True
        # will be taken into consideration.
        # If `update_stats` is given as True and observation normalization is enabled, then we will update the
        # RunningNorm instances as well.

        if self._observation_normalization:
            # This is the case where observation normalization is enabled.
            if self._obs_stats is None:
                # If we do not have observation stats yet, we build a new one (according to the dtype and device
                # of the observation).
                self._obs_stats = self._make_running_norm(observation)
            else:
                # If we already have observation stats, we make sure that it is in the correct device.
                self._obs_stats = self._transfer_running_norm(self._obs_stats, observation)

            if update_stats:
                # This is the case where the `update_stats` argument was encountered as True.
                if self._collected_stats is None:
                    # If the RunningNorm responsible to collect new stats is not built yet, we build it here
                    # (according to the dtype and device of the observation).
                    self._collected_stats = self._make_running_norm(observation)
                else:
                    # If the RunningNorm responsible to collect new stats already exists, then we make sure
                    # that it is in the correct device.
                    self._collected_stats = self._transfer_running_norm(self._collected_stats, observation)

                # We first update the RunningNorm responsible for collecting the new stats.
                self._collected_stats.update(observation, mask)

                # We now update the RunningNorm which stores all the stats, and return the normalized observation.
                result = self._obs_stats.update_and_normalize(observation, mask)
            else:
                # This is the case where the `update_stats` argument was encountered as False.
                # Here we normalize the observation but do not update our existing RunningNorm instances.
                result = self._obs_stats.update(observation, mask)
            return result
        else:
            # This is the case where observation normalization is disabled.
            # In this case, we just return the observation as it is.
            return observation

    def _ensure_obsnorm(self):
        if not self.observation_normalization:
            raise ValueError("This feature can only be used when observation_normalization=True.")

    def get_observation_stats(self) -> RunningNorm:
        """Get the observation stats"""
        self._ensure_obsnorm()
        return self._obs_stats

    def _make_sync_data_for_actors(self) -> Any:
        if self.observation_normalization:
            obs_stats = self.get_observation_stats()
            if obs_stats is not None:
                obs_stats = obs_stats.to("cpu")
            return dict(obs_stats=obs_stats)
        else:
            return None

    def set_observation_stats(self, rn: RunningNorm):
        """Set the observation stats"""
        self._ensure_obsnorm()
        self._obs_stats = rn

    def _use_sync_data_from_main(self, received: dict):
        for k, v in received.items():
            if k == "obs_stats":
                self.set_observation_stats(v)

    def pop_observation_stats(self) -> RunningNorm:
        """Get and clear the collected observation stats"""
        self._ensure_obsnorm()
        result = self._collected_stats
        self._collected_stats = None
        return result

    def _make_sync_data_for_main(self) -> Any:
        result = dict(episode_count=self.episode_count, interaction_count=self.interaction_count)

        if self.observation_normalization:
            collected = self.pop_observation_stats()
            if collected is not None:
                collected = collected.to("cpu")
            result["obs_stats_delta"] = collected

        return result

    def update_observation_stats(self, rn: RunningNorm):
        """Update the observation stats via another RunningNorm instance"""
        self._ensure_obsnorm()
        if self._obs_stats is None:
            self._obs_stats = rn
        else:
            self._obs_stats.update(rn)

    def _use_sync_data_from_actors(self, received: list):
        total_episode_count = 0
        total_interaction_count = 0

        for data in received:
            data: dict
            total_episode_count += data["episode_count"]
            total_interaction_count += data["interaction_count"]
            if self.observation_normalization:
                self.update_observation_stats(data["obs_stats_delta"])

        self.set_episode_count(total_episode_count)
        self.set_interaction_count(total_interaction_count)

    def _make_pickle_data_for_main(self) -> dict:
        # For when the main Problem object (the non-remote one) gets pickled,
        # this function returns the counters of this remote Problem instance,
        # to be sent to the main one.
        return dict(interaction_count=self.interaction_count, episode_count=self.episode_count)

    def _use_pickle_data_from_main(self, state: dict):
        # For when a newly unpickled Problem object gets (re)parallelized,
        # this function restores the inner states specific to this remote
        # worker. In the case of GymNE, those inner states are episode
        # and interaction counters.
        for k, v in state.items():
            if k == "episode_count":
                self.set_episode_count(v)
            elif k == "interaction_count":
                self.set_interaction_count(v)
            else:
                raise ValueError(f"When restoring the inner state of a remote worker, unrecognized state key: {k}")

    def _evaluate_batch(self, batch: SolutionBatch):
        if self._actor_max_num_envs is None:
            self._evaluate_subbatch(batch)
        else:
            subbatches = batch.split(max_size=self._actor_max_num_envs)
            for subbatch in subbatches:
                self._evaluate_subbatch(subbatch)

    def _evaluate_subbatch(self, batch: SolutionBatch):
        # Get the number of solutions and the solution batch from the shape of the batch.
        num_solutions, solution_length = batch.values_shape

        # Get (possibly after (re)building) the environment object.
        env = self._get_env(num_solutions)

        # Reset the environment and receive the first observation batch.
        obs_per_env = env.reset()

        # Update the simulator device according to the device of the observation batch received.
        self._simulator_device = obs_per_env.device

        # Get the number of environments.
        num_envs = obs_per_env.shape[0]

        # Transfer (if necessary) the solutions (which are the network parameters) to the simulator device.
        batch_values = batch.values.to(self._simulator_device)

        if num_solutions == num_envs:
            # If the number of solutions is equal to the number of environments, then we declare all of the solutions
            # as the network parameters, and we declare all of these environments active.
            params_per_env = batch_values
            active_per_env = torch.ones(num_solutions, dtype=torch.bool, device=self._simulator_device)
        elif num_solutions < num_envs:
            # If the number of solutions is less than the number of environments, then we allocate a new empty
            # tensor to represent the network parameters.
            params_per_env = torch.empty((num_envs, solution_length), dtype=batch.dtype, device=self._simulator_device)

            # The first `num_solutions` rows of this new parameters tensor is filled with the values of the solutions.
            params_per_env[:num_solutions, :] = batch_values

            # The remaining parameters become the clones of the first solution.
            params_per_env[num_solutions:, :] = batch_values[0]

            # At first, all the environments are declared as inactive.
            active_per_env = torch.zeros(num_envs, dtype=torch.bool, device=self._simulator_device)

            # Now, the first `num_solutions` amount of environments is declared as active.
            # The remaining ones remain inactive.
            active_per_env[:num_solutions] = True
        else:
            assert False, "Received incompatible number of environments"

        # We get the policy and fill it with the parameters stored by the solutions.
        policy = self._policy
        policy.set_parameters(params_per_env)

        # Declare the counter which stores the total timesteps encountered during this evaluation.
        total_timesteps = 0

        # Declare the counters (one for each environment) storing the number of episodes completed.
        num_eps_per_env = torch.zeros(num_envs, dtype=torch.int64, device=self._simulator_device)

        # Declare the scores (one for each environment).
        score_per_env = torch.zeros(num_envs, dtype=torch.float32, device=self._simulator_device)

        if self._alive_bonus_schedule is not None:
            # If an alive_bonus_schedule was provided, then we extract the timesteps.
            # bonus_t0 is the timestep where the partial alive bonus will start.
            # bonus_t1 is the timestep where the full alive bonus will start.
            # alive_bonus is the amount that will be added to reward if the agent is alive.
            bonus_t0, bonus_t1, alive_bonus = self._alive_bonus_schedule

            if bonus_t1 > bonus_t0:
                # If bonus_t1 is bigger than bonus_t0, then we have a partial alive bonus time window.
                add_partial_alive_bonus = True

                # We compute and store the length of the time window.
                bonus_t_gap_as_float = float(bonus_t1 - bonus_t0)
            else:
                # If bonus_t1 is NOT bigger than bonus_t0, then we do NOT have a partial alive bonus time window.
                add_partial_alive_bonus = False

            # To properly give the alive bonus for each solution, we need to keep track of the timesteps for all
            # the running solutions. So, we declare the following variable.
            t_per_env = torch.zeros(num_envs, dtype=torch.int64, device=self._simulator_device)

        # We normalize the initial observation.
        obs_per_env = self._normalize_observation(obs_per_env, mask=active_per_env)

        while True:
            # Pass the observations through the policy and get the actions to perform.
            action_per_env = policy(torch.as_tensor(obs_per_env, dtype=params_per_env.dtype))

            if self._action_noise_stdev is not None:
                # If we are to apply action noise, we sample from a Gaussian distribution and add the noise onto
                # the actions.
                action_per_env = action_per_env + (torch.rand_like(action_per_env) * self._action_noise_stdev)

            # Apply the actions, get the observations, rewards, and the 'done' flags.
            obs_per_env, reward_per_env, done_per_env, _ = env.step(action_per_env)

            if self._decrease_rewards_by is not None:
                # We decrease the rewards, if we have the configuration to do so.
                reward_per_env = reward_per_env - self._decrease_rewards_by

            if self._alive_bonus_schedule is not None:
                # Here we handle the alive bonus schedule.

                # For each environment, increment the timestep.
                t_per_env[active_per_env] += 1

                # For those who are within the full alive bonus time region, increase the scores by the full amount.
                in_full_bonus_t_per_env = active_per_env & (t_per_env >= bonus_t1)
                score_per_env[in_full_bonus_t_per_env] += alive_bonus

                if add_partial_alive_bonus:
                    # Here we handle the partial alive bonus time window.
                    # We first determine which environments are in the partial alive bonus time window.
                    in_partial_bonus_t_per_env = active_per_env & (t_per_env >= bonus_t0) & (t_per_env < bonus_t1)

                    # Here we compute the partial alive bonuses and add those bonuses to the scores.
                    score_per_env[in_partial_bonus_t_per_env] += alive_bonus * (
                        torch.as_tensor(t_per_env[in_partial_bonus_t_per_env] - bonus_t0, dtype=torch.float32)
                        / bonus_t_gap_as_float
                    )

                # Determine which environments just finished their episodes.
                just_finished_per_env = active_per_env & done_per_env

                # Reset the timestep counters of the environments which are just finished.
                t_per_env[just_finished_per_env] = 0

            # For each active environment, increase the score by the reward received.
            score_per_env[active_per_env] += reward_per_env[active_per_env]

            # Update the total timesteps counter.
            total_timesteps += int(torch.sum(active_per_env))

            # Reset the policies whose episodes are done (so that their hidden states become 0).
            policy.reset(done_per_env)

            # Update the number of episodes counter for each environment.
            num_eps_per_env[done_per_env] += 1

            # Solutions with number of completed episodes larger than the number of allowed episodes become inactive.
            active_per_env[:num_solutions] = num_eps_per_env[:num_solutions] < self._num_episodes

            if not torch.any(active_per_env[:num_solutions]):
                # If there is not a single active solution left, then we exit this loop.
                break

            # For the next iteration of this loop, we normalize the observation.
            obs_per_env = self._normalize_observation(obs_per_env, mask=active_per_env)

        # Update the interaction count and the episode count stored by this VecGymNE instance.
        self._interaction_count += total_timesteps
        self._episode_count += num_solutions * self._num_episodes

        # Compute the fitnesses
        fitnesses = score_per_env[:num_solutions]
        if self._num_episodes > 1:
            fitnesses /= self._num_episodes

        # Assign the scores to the solutions as fitnesses.
        batch.set_evals(fitnesses)

    def get_env(self) -> Optional[gym.Env]:
        """
        Get the gym environment.

        Returns:
            The gym environment if it is built. If not built yet, None.
        """
        return self._env

    def to_policy(self, solution: Iterable, *, with_wrapper_modules: bool = True) -> nn.Module:
        """
        Convert the given solution to a policy.

        Args:
            solution: A solution which can be given as a `torch.Tensor`, as a
                `Solution`, or as any `Iterable`.
            with_wrapper_modules: Whether or not to wrap the policy module
                with helper modules so that observations are normalized
                and actions are clipped to be within the correct boundaries.
                The default and the recommended value is True.
        Returns:
            The policy, as a `torch.nn.Module` instance.
        """
        # Get the gym environment
        env = self._get_env(1)

        # Get the observation space, its lower and higher bounds.
        obs_space = env.single_action_space
        low = obs_space.low
        high = obs_space.high

        # If the lower and higher bounds are not -inf and +inf respectively, then this environment needs clipping.
        needs_clipping = _numpy_arrays_specify_bounds(low, high)

        # Convert the solution to a PyTorch tensor on cpu.
        if isinstance(solution, torch.Tensor):
            solution = solution.to("cpu")
        elif isinstance(solution, Solution):
            solution = solution.values.clone().to("cpu")
        else:
            solution = torch.as_tensor(solution, dtype=torch.float32, device="cpu")

        # Convert the internally stored policy to a PyTorch module.
        result = self._policy.to_torch_module(solution)

        if with_wrapper_modules:
            if self.observation_normalization and (self._obs_stats is not None):
                # If observation normalization is needed and there are collected observation stats, then we wrap the
                # policy with an ObsNormWrapperModule.
                result = ObsNormWrapperModule(result, self._obs_stats)

            if needs_clipping:
                # If clipping is needed, then we wrap the policy with an ActClipWrapperModule
                result = ActClipWrapperModule(result, obs_space)

        return result

    def save_solution(self, solution: Iterable, fname: Union[str, Path]):
        """
        Save the solution into a pickle file.
        Among the saved data within the pickle file are the solution
        (as a PyTorch tensor), the policy (as a `torch.nn.Module` instance),
        and observation stats (if any).

        Args:
            solution: The solution to be saved. This can be a PyTorch tensor,
                a `Solution` instance, or any `Iterable`.
            fname: The file name of the pickle file to be created.
        """

        # Convert the solution to a PyTorch tensor on the cpu.
        if isinstance(solution, torch.Tensor):
            solution = solution.to("cpu")
        elif isinstance(solution, Solution):
            solution = solution.values.clone().to("cpu")
        else:
            solution = torch.as_tensor(solution, dtype=torch.float32, device="cpu")

        if isinstance(solution, ReadOnlyTensor):
            solution = solution.as_subclass(torch.Tensor)

        # Store the solution and the policy.
        result = {
            "solution": solution,
            "policy": self.to_policy(solution),
        }

        # If available, store the observation stats.
        if self.observation_normalization and (self._obs_stats is not None):
            result["obs_mean"] = self._obs_stats.mean.to("cpu")
            result["obs_stdev"] = self._obs_stats.stdev.to("cpu")
            result["obs_sum"] = self._obs_stats.sum.to("cpu")
            result["obs_sum_of_squares"] = self._obs_stats.sum_of_squares.to("cpu")

        # Some additional data.
        result["interaction_count"] = self.interaction_count
        result["episode_count"] = self.episode_count
        result["time"] = datetime.now()

        if isinstance(self._env_maker, str):
            # If the environment was specified via a string, store the string.
            result["env"] = self._env_maker

        # Store the network architecture.
        result["architecture"] = self._architecture

        # Save the dictionary which stores the data.
        with open(fname, "wb") as f:
            pickle.dump(result, f)

    @property
    def max_num_envs(self) -> Optional[int]:
        """
        Maximum number of environments to be allocated.

        If a maximum number of environments is not set, then None is returned.
        If this problem instance is the main one, then the overall maximum
        number of environments is returned.
        If this problem instance is a remote one (i.e. is on a remote actor)
        then the maximum number of environments for that actor is returned.
        """
        if self.is_main:
            return self._max_num_envs
        else:
            return self._actor_max_num_envs

    def make_net(self, solution: Iterable) -> nn.Module:
        """
        Make a new policy network parameterized by the given solution.
        Note that this parameterized network assumes that the observation
        is already normalized, and it does not do action clipping to ensure
        that the generated actions are within valid bounds.

        To have a policy network which has its own observation normalization
        and action clipping layers, please see the method `to_policy(...)`.

        Args:
            solution: The solution which stores the parameters.
                This can be a Solution instance, or a 1-dimensional tensor,
                or any Iterable of real numbers.
        Returns:
            The policy network, as a PyTorch module.
        """
        return self.to_policy(solution, with_wrapper_modules=False)

    @property
    def network_device(self) -> Optional[Device]:
        """
        The device on which the policy networks will operate.

        Specific to VecGymNE, the network device is determined only
        after receiving the first observation from the reinforcement
        learning environment. Until then, this property has the value
        None.
        """
        return self._simulator_device



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various neural network layer types"""

from typing import Callable, List, Optional, Tuple, Union

import torch
import torch.nn.functional as nnf
from torch import nn


class Clip(nn.Module):
    """A small torch module for clipping the values of tensors"""

    def __init__(self, lb: float, ub: float):
        """`__init__(...)`: Initialize the Clip operator.

        Args:
            lb: Lower bound. Values less than this will be clipped.
            ub: Upper bound. Values greater than this will be clipped.
        """
        nn.Module.__init__(self)
        self._lb = float(lb)
        self._ub = float(ub)

    def forward(self, x: torch.Tensor):
        return x.clamp(self._lb, self._ub)

    def extra_repr(self):
        return "lb={}, ub={}".format(self._lb, self._ub)


class Bin(nn.Module):
    """A small torch module for binning the values of tensors.

    In more details, considering a lower bound value lb,
    an upper bound value ub, and an input tensor x,
    each value within x closer to lb will be converted to lb
    and each value within x closer to ub will be converted to ub.
    """

    def __init__(self, lb: float, ub: float):
        """`__init__(...)`: Initialize the Clip operator.

        Args:
            lb: Lower bound
            ub: Upper bound
        """
        nn.Module.__init__(self)
        self._lb = float(lb)
        self._ub = float(ub)
        self._interval_size = self._ub - self._lb
        self._shrink_amount = self._interval_size / 2.0
        self._shift_amount = (self._ub + self._lb) / 2.0

    def forward(self, x: torch.Tensor):
        x = x - self._shift_amount
        x = x / self._shrink_amount
        x = torch.sign(x)
        x = x * self._shrink_amount
        x = x + self._shift_amount
        return x

    def extra_repr(self):
        return "lb={}, ub={}".format(self._lb, self._ub)


class Slice(nn.Module):
    """A small torch module for getting the slice of an input tensor"""

    def __init__(self, from_index: int, to_index: int):
        """`__init__(...)`: Initialize the Slice operator.

        Args:
            from_index: The index from which the slice begins.
            to_index: The exclusive index at which the slice ends.
        """
        nn.Module.__init__(self)
        self._from_index = from_index
        self._to_index = to_index

    def forward(self, x):
        return x[self._from_index : self._to_index]

    def extra_repr(self):
        return "from_index={}, to_index={}".format(self._from_index, self._to_index)


class Round(nn.Module):
    """A small torch module for rounding the values of an input tensor"""

    def __init__(self, ndigits: int = 0):
        nn.Module.__init__(self)
        self._ndigits = int(ndigits)
        self._q = 10.0**self._ndigits

    def forward(self, x):
        x = x * self._q
        x = torch.round(x)
        x = x / self._q
        return x

    def extra_repr(self):
        return "ndigits=" + str(self._ndigits)


class Apply(nn.Module):
    """A torch module for applying an arithmetic operator on an input tensor"""

    def __init__(self, operator: str, argument: float):
        """`__init__(...)`: Initialize the Apply module.

        Args:
            operator: Must be '+', '-', '*', '/', or '**'.
                Indicates which operation will be done
                on the input tensor.
            argument: Expected as a float, represents
                the right-argument of the operation
                (the left-argument being the input
                tensor).
        """
        nn.Module.__init__(self)

        self._operator = str(operator)
        assert self._operator in ("+", "-", "*", "/", "**")

        self._argument = float(argument)

    def forward(self, x):
        op = self._operator
        arg = self._argument
        if op == "+":
            return x + arg
        elif op == "-":
            return x - arg
        elif op == "*":
            return x * arg
        elif op == "/":
            return x / arg
        elif op == "**":
            return x**arg
        else:
            raise ValueError("Unknown operator:" + repr(op))

    def extra_repr(self):
        return "operator={}, argument={}".format(repr(self._operator), self._argument)


class RNN(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        nonlinearity: str = "tanh",
        *,
        dtype: torch.dtype = torch.float32,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__()

        input_size = int(input_size)
        hidden_size = int(hidden_size)
        nonlinearity = str(nonlinearity)

        self.W1 = nn.Parameter(torch.randn(hidden_size, input_size, dtype=dtype, device=device))
        self.W2 = nn.Parameter(torch.randn(hidden_size, hidden_size, dtype=dtype, device=device))
        self.b1 = nn.Parameter(torch.zeros(hidden_size, dtype=dtype, device=device))
        self.b2 = nn.Parameter(torch.zeros(hidden_size, dtype=dtype, device=device))

        if nonlinearity == "tanh":
            self.actfunc = torch.tanh
        else:
            self.actfunc = getattr(nnf, nonlinearity)

        self.nonlinearity = nonlinearity
        self.input_size = input_size
        self.hidden_size = hidden_size

    def forward(self, x: torch.Tensor, h: Optional[torch.Tensor] = None) -> tuple:
        if h is None:
            h = torch.zeros(self.hidden_size, dtype=x.dtype, device=x.device)
        act = self.actfunc
        W1 = self.W1
        W2 = self.W2
        b1 = self.b1.unsqueeze(-1)
        b2 = self.b2.unsqueeze(-1)
        x = x.unsqueeze(-1)
        h = h.unsqueeze(-1)
        y = act(((W1 @ x) + b1) + ((W2 @ h) + b2))
        y = y.squeeze(-1)
        return y, y

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"{clsname}(input_size={self.input_size}, hidden_size={self.hidden_size}, nonlinearity={repr(self.nonlinearity)})"


class LSTM(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        *,
        dtype: torch.dtype = torch.float32,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__()
        input_size = int(input_size)
        hidden_size = int(hidden_size)

        self.input_size = input_size
        self.hidden_size = hidden_size

        def input_weight():
            return nn.Parameter(torch.randn(self.hidden_size, self.input_size, dtype=dtype, device=device))

        def weight():
            return nn.Parameter(torch.randn(self.hidden_size, self.hidden_size, dtype=dtype, device=device))

        def bias():
            return nn.Parameter(torch.zeros(self.hidden_size, dtype=dtype, device=device))

        self.W_ii = input_weight()
        self.W_if = input_weight()
        self.W_ig = input_weight()
        self.W_io = input_weight()

        self.W_hi = weight()
        self.W_hf = weight()
        self.W_hg = weight()
        self.W_ho = weight()

        self.b_ii = bias()
        self.b_if = bias()
        self.b_ig = bias()
        self.b_io = bias()

        self.b_hi = bias()
        self.b_hf = bias()
        self.b_hg = bias()
        self.b_ho = bias()

    def forward(self, x: torch.Tensor, hidden=None) -> tuple:
        sigm = torch.sigmoid
        tanh = torch.tanh

        if hidden is None:
            h_prev = torch.zeros(self.hidden_size, dtype=x.dtype, device=x.device)
            c_prev = torch.zeros(self.hidden_size, dtype=x.dtype, device=x.device)
        else:
            h_prev, c_prev = hidden

        i_t = sigm(self.W_ii @ x + self.b_ii + self.W_hi @ h_prev + self.b_hi)
        f_t = sigm(self.W_if @ x + self.b_if + self.W_hf @ h_prev + self.b_hf)
        g_t = tanh(self.W_ig @ x + self.b_ig + self.W_hg @ h_prev + self.b_hg)
        o_t = sigm(self.W_io @ x + self.b_io + self.W_ho @ h_prev + self.b_ho)
        c_t = f_t * c_prev + i_t * g_t
        h_t = o_t * tanh(c_t)

        return h_t, (h_t, c_t)

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"{clsname}(input_size={self.input_size}, hidden_size={self.hidden_size})"


RecurrentNet = RNN
LSTMNet = LSTM


class FeedForwardNet(nn.Module):
    """
    Representation of a feed forward neural network as a torch Module.

    An example initialization of a FeedForwardNet is as follows:

        net = drt.FeedForwardNet(4, [(8, 'tanh'), (6, 'tanh')])

    which means that we would like to have a network which expects an input
    vector of length 4 and passes its input through 2 tanh-activated hidden
    layers (with neurons count 8 and 6, respectively).
    The output of the last hidden layer (of length 6) is the final
    output vector.

    The string representation of the module obtained via the example above
    is:

        FeedForwardNet(
          (layer_0): Linear(in_features=4, out_features=8, bias=True)
          (actfunc_0): Tanh()
          (layer_1): Linear(in_features=8, out_features=6, bias=True)
          (actfunc_1): Tanh()
        )
    """

    LengthActTuple = Tuple[int, Union[str, Callable]]
    LengthActBiasTuple = Tuple[int, Union[str, Callable], Union[bool]]

    def __init__(self, input_size: int, layers: List[Union[LengthActTuple, LengthActBiasTuple]]):
        """`__init__(...)`: Initialize the FeedForward network.

        Args:
            input_size: Input size of the network, expected as an int.
            layers: Expected as a list of tuples,
                where each tuple is either of the form
                `(layer_size, activation_function)`
                or of the form
                `(layer_size, activation_function, bias)`
                in which
                (i) `layer_size` is an int, specifying the number of neurons;
                (ii) `activation_function` is None, or a callable object,
                or a string containing the name of the activation function
                ('relu', 'selu', 'elu', 'tanh', 'hardtanh', or 'sigmoid');
                (iii) `bias` is a boolean, specifying whether the layer
                is to have a bias or not.
                When omitted, bias is set to True.
        """

        nn.Module.__init__(self)

        for i, layer in enumerate(layers):
            if len(layer) == 2:
                size, actfunc = layer
                bias = True
            elif len(layer) == 3:
                size, actfunc, bias = layer
            else:
                assert False, "A layer tuple of invalid size is encountered"

            setattr(self, "layer_" + str(i), nn.Linear(input_size, size, bias=bias))

            if isinstance(actfunc, str):
                if actfunc == "relu":
                    actfunc = nn.ReLU()
                elif actfunc == "selu":
                    actfunc = nn.SELU()
                elif actfunc == "elu":
                    actfunc = nn.ELU()
                elif actfunc == "tanh":
                    actfunc = nn.Tanh()
                elif actfunc == "hardtanh":
                    actfunc = nn.Hardtanh()
                elif actfunc == "sigmoid":
                    actfunc = nn.Sigmoid()
                elif actfunc == "round":
                    actfunc = Round()
                else:
                    raise ValueError("Unknown activation function: " + repr(actfunc))

            setattr(self, "actfunc_" + str(i), actfunc)

            input_size = size

    def forward(self, x):
        i = 0
        while hasattr(self, "layer_" + str(i)):
            x = getattr(self, "layer_" + str(i))(x)
            f = getattr(self, "actfunc_" + str(i))
            if f is not None:
                x = f(x)
            i += 1
        return x


class StructuredControlNet(nn.Module):
    """Structured Control Net.

    This is a control network consisting of two components:
    (i) a non-linear component, which is a feed-forward network; and
    (ii) a linear component, which is a linear layer.
    Both components take the input vector provided to the
    structured control network.
    The final output is the sum of the outputs of both components.

    Reference:
        Mario Srouji, Jian Zhang, Ruslan Salakhutdinov (2018).
        Structured Control Nets for Deep Reinforcement Learning.
    """

    def __init__(
        self,
        *,
        in_features: int,
        out_features: int,
        num_layers: int,
        hidden_size: int,
        bias: bool = True,
        nonlinearity: Union[str, Callable] = "tanh",
    ):
        """`__init__(...)`: Initialize the structured control net.

        Args:
            in_features: Length of the input vector
            out_features: Length of the output vector
            num_layers: Number of hidden layers for the non-linear component
            hidden_size: Number of neurons in a hidden layer of the
                non-linear component
            bias: Whether or not the linear component is to have bias
            nonlinearity: Activation function
        """

        nn.Module.__init__(self)

        self._in_features = in_features
        self._out_features = out_features
        self._num_layers = num_layers
        self._hidden_size = hidden_size
        self._bias = bias
        self._nonlinearity = nonlinearity

        self._linear_component = nn.Linear(
            in_features=self._in_features, out_features=self._out_features, bias=self._bias
        )

        self._nonlinear_component = FeedForwardNet(
            input_size=self._in_features,
            layers=(
                list((self._hidden_size, self._nonlinearity) for _ in range(self._num_layers))
                + [(self._out_features, self._nonlinearity)]
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """TODO: documentation"""
        return self._linear_component(x) + self._nonlinear_component(x)

    @property
    def in_features(self):
        """TODO: documentation"""
        return self._in_features

    @property
    def out_features(self):
        """TODO: documentation"""
        return self._out_features

    @property
    def num_layers(self):
        """TODO: documentation"""
        return self._num_layers

    @property
    def hidden_size(self):
        """TODO: documentation"""
        return self._hidden_size

    @property
    def bias(self):
        """TODO: documentation"""
        return self._bias

    @property
    def nonlinearity(self):
        """TODO: documentation"""
        return self._nonlinearity


class LocomotorNet(nn.Module):
    """LocomotorNet: A locomotion-specific structured control net.

    This is a control network which consists of two components:
    one linear, and one non-linear. The non-linear component
    is an input-independent set of sinusoidals waves whose
    amplitudes, frequencies and phases are trainable.
    Upon execution of a forward pass, the output of the non-linear
    component is the sum of all these sinusoidal waves.
    The linear component is a linear layer (optionally with bias)
    whose weights (and biases) are trainable.
    The final output of the LocomotorNet at the end of a forward pass
    is the sum of the linear and the non-linear components.

    Note that this is a stateful network, where the only state
    is the timestep t, which starts from 0 and gets incremented by 1
    at the end of each forward pass. The `reset()` method resets
    t back to 0.

    Reference:
        Mario Srouji, Jian Zhang, Ruslan Salakhutdinov (2018).
        Structured Control Nets for Deep Reinforcement Learning.
    """

    def __init__(self, *, in_features: int, out_features: int, bias: bool = True, num_sinusoids=16):
        """`__init__(...)`: Initialize the LocomotorNet.

        Args:
            in_features: Length of the input vector
            out_features: Length of the output vector
            bias: Whether or not the linear component is to have a bias
            num_sinusoids: Number of sinusoidal waves
        """

        nn.Module.__init__(self)

        self._in_features = in_features
        self._out_features = out_features
        self._bias = bias
        self._num_sinusoids = num_sinusoids

        self._linear_component = nn.Linear(
            in_features=self._in_features, out_features=self._out_features, bias=self._bias
        )

        self._amplitudes = nn.ParameterList()
        self._frequencies = nn.ParameterList()
        self._phases = nn.ParameterList()

        for _ in range(self._num_sinusoids):
            for paramlist in (self._amplitudes, self._frequencies, self._phases):
                paramlist.append(nn.Parameter(torch.randn(self._out_features, dtype=torch.float32)))

        self.reset()

    def reset(self):
        """Set the timestep t to 0"""
        self._t = 0

    @property
    def t(self) -> int:
        """The current timestep t"""
        return self._t

    @property
    def in_features(self) -> int:
        """Get the length of the input vector"""
        return self._in_features

    @property
    def out_features(self) -> int:
        """Get the length of the output vector"""
        return self._out_features

    @property
    def num_sinusoids(self) -> int:
        """Get the number of sinusoidal waves of the non-linear component"""
        return self._num_sinusoids

    @property
    def bias(self) -> bool:
        """Get whether or not the linear component has bias"""
        return self._bias

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Execute a forward pass"""
        u_linear = self._linear_component(x)

        t = self._t
        u_nonlinear = torch.zeros(self._out_features)
        for i in range(self._num_sinusoids):
            A = self._amplitudes[i]
            w = self._frequencies[i]
            phi = self._phases[i]
            u_nonlinear = u_nonlinear + (A * torch.sin(w * t + phi))

        self._t += 1

        return u_linear + u_nonlinear



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import pytest
import torch
from torch import nn

from evotorch.neuroevolution.net import (
    MultiLayered,
    fill_parameters,
    layers,
    make_functional_module,
    parameter_vector,
    str_to_net,
)
from evotorch.testing import assert_allclose


class Unbatched:
    # These are wrappers around `torch.nn.LSTM` and `torch.nn.RNN`.
    # The interface is simplified so that the inputs, outputs, and the hidden states have neither a batch dimension
    # nor a timestep dimension.

    class LSTM(nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            self.lstm = nn.LSTM(*args, **kwargs)

        def forward(self, x: torch.Tensor, h: Optional[tuple] = None) -> tuple:
            if h is not None:
                a, b = h
                a = a.reshape(1, 1, -1)
                b = b.reshape(1, 1, -1)
                h = a, b

            x = x.reshape(1, 1, -1)
            x, h = self.lstm(x, h)
            x = x.reshape(-1)

            a, b = h
            a = a.reshape(-1)
            b = b.reshape(-1)
            h = a, b

            return x, h

    class RNN(nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            self.rnn = nn.RNN(*args, **kwargs)

        def forward(self, x: torch.Tensor, h: Optional[torch.Tensor] = None) -> tuple:
            if h is not None:
                h = h.reshape(1, 1, -1)
            x = x.reshape(1, 1, -1)
            x, h = self.rnn(x, h)
            x = x.reshape(-1)
            h = h.reshape(-1)
            return x, h


@pytest.mark.parametrize("class_name", ["LSTM", "RNN"])
def test_recurrent_network(class_name: str):
    tolerance = 1e-2

    # Randomly generate some inputs
    inputs = torch.randn(4, 3)

    # Get the unbatched version of the recurrent network from `torch.nn`
    net_from_unbatched = getattr(Unbatched, class_name)(3, 5)

    # Get the EvoTorch counterpart of the recurrent network.
    net_from_layers = getattr(layers, class_name)(3, 5)

    # Get the parameter vector from the first network and fill the parameters of the second network using its values.
    nn_params = parameter_vector(net_from_unbatched)
    fill_parameters(net_from_layers, nn_params)

    # Now that both networks have the same parameters, they should generate the same outputs.
    # We now feed both networks the same inputs.
    with torch.no_grad():
        h_of_unbatched_net = None
        h_of_layers_net = None
        for x in inputs:
            y_of_unbatched_net, h_of_unbatched_net = net_from_unbatched(x, h_of_unbatched_net)
            y_of_layers_net, h_of_layers_net = net_from_layers(x, h_of_layers_net)

    # Check if both networks report the same output.
    assert_allclose(y_of_layers_net, y_of_unbatched_net, atol=tolerance)


def test_multilayered():
    # When recurrent networks are NOT involved, a MultiLayered module should behave like a `nn.Sequential instance`.
    # Here, we instantiate an `nn.Sequential` and a MultiLayered, and fill them with the same layers.
    tolerance = 1e-2

    layer1 = nn.Linear(3, 5)
    layer2 = nn.Linear(5, 2)

    net1 = nn.Sequential(layer1, nn.Tanh(), layer2)
    net2 = MultiLayered(layer1, nn.Tanh(), layer2)

    x = torch.rand(3)

    with torch.no_grad():
        y1 = net1(x)
        y2 = net2(x)

    # Given the same inputs, these two networks should generate the same output.
    assert_allclose(y2, y1, atol=tolerance)


class DummyRecurrentNet(nn.Module):
    def __init__(self, first_value: int = 1):
        super().__init__()
        self.first_value = int(first_value)

    def forward(self, x: torch.Tensor, h: Optional[torch.Tensor] = None) -> tuple:
        if h is None:
            h = torch.tensor(self.first_value, dtype=torch.int64, device=x.device)
        return x * torch.as_tensor(h, dtype=x.dtype, device=x.device), h + 1


@torch.no_grad()
def test_multilayered_with_recurrence():
    # When it contains recurrent layers, a MultiLayered stores the hidden states of its layers in a dictionary,
    # and then returns that dictionary as its own hidden state.
    # In this test, we instantiate a MultiLayered with two recurrent layers, and then check if its generates
    # its outputs correctly after receiving two inputs.
    tolerance = 1e-3

    net = MultiLayered(DummyRecurrentNet(first_value=1), nn.ReLU(), DummyRecurrentNet(first_value=2))

    x = torch.tensor([-1.0, 2.0, 3.0], dtype=torch.float32)

    x, h = net(x)
    assert_allclose(x, [0.0, 4.0, 6.0], atol=tolerance)
    assert isinstance(h, dict)
    assert set(h.keys()) == {0, 2}
    assert int(h[0]) == 2
    assert int(h[2]) == 3

    x, h = net(x, h)
    assert_allclose(x, [0.0, 24.0, 36.0], atol=tolerance)
    assert isinstance(h, dict)
    assert set(h.keys()) == {0, 2}
    assert int(h[0]) == 3
    assert int(h[2]) == 4


@torch.no_grad()
def test_str_to_net():
    tolerance = 1e-2

    # Make an `nn.Sequential` instance, and an equivalent network using `str_to_net(...)`.
    net = nn.Sequential(nn.Linear(3, 5), nn.Tanh(), nn.Linear(5, 2))
    snet = str_to_net("Linear(3, 5) >> Tanh() >> Linear(5, 2)")

    # Make sure that both networks are using the same parameters.
    params = parameter_vector(net)
    fill_parameters(snet, params)

    # Generate random inputs
    x = torch.randn(3)

    # Feed both networks the same inputs
    y = net(x)
    sy = snet(x)

    # Assert that the outputs match
    assert_allclose(sy, y, rtol=tolerance)


class DummyComposedRecurrent(nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = nn.ModuleList([Unbatched.RNN(3, 5), nn.Linear(5, 8), Unbatched.LSTM(8, 2)])

    def forward(self, x: torch.Tensor, h: Optional[dict] = None) -> tuple:
        if h is None:
            h = {0: None, 2: None}

        x, h[0] = self.layers[0](x, h[0])
        x = self.layers[1](x)
        x = torch.tanh(x)
        x, h[2] = self.layers[2](x, h[2])

        return x, h


@torch.no_grad()
def test_str_to_net_with_recurrence():
    tolerance = 1e-2

    net = DummyComposedRecurrent()
    snet = str_to_net("RNN(3, 5) >> Linear(5, 8) >> Tanh() >> LSTM(8, 2)")

    params = parameter_vector(net)
    fill_parameters(snet, params)

    inputs = torch.randn(10, 3)

    h = None
    sh = None
    for x in inputs:
        y, h = net(x, h)
        sy, sh = snet(x, sh)

    assert_allclose(sy, y, atol=tolerance)


@torch.no_grad()
def test_make_functional_module():
    tolerance = 1e-2

    net = str_to_net("Linear(3, 8) >> Tanh() >> Linear(8, 2)")
    fnet = make_functional_module(net)

    params = parameter_vector(net)

    x = torch.randn(3)
    y1 = net(x)
    y2 = fnet(params, x)

    assert_allclose(y2, y1, atol=tolerance)


@torch.no_grad()
def test_make_functional_module_with_lstm():
    tolerance = 1e-2

    net = str_to_net("LSTM(3, 8) >> Linear(8, 2)")
    fnet = make_functional_module(net)

    params = parameter_vector(net)

    inputs = torch.randn(7, 3)

    h1 = None
    h2 = None
    for x in inputs:
        y1, h1 = net(x, h1)
        y2, h2 = fnet(params, x, h2)

    assert_allclose(y2, y1, atol=tolerance)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from torch import nn

from evotorch.neuroevolution.net import fill_parameters, parameter_vector
from evotorch.neuroevolution.net.layers import LSTM, RNN
from evotorch.neuroevolution.net.vecrl import Policy, reset_tensors
from evotorch.testing import assert_allclose
from evotorch.tools import clone


def test_reset_tensors_simple():
    tolerance = 1e-4

    # Make a two-dimensional tensor
    x = torch.tensor(
        [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11],
            [12, 13, 14, 15],
        ],
        dtype=torch.float32,
    )

    # Use `reset_tensors(...)` to reset the rows with indices 0 and 2
    reset_tensors(x, [0, 2])

    # Assert that the rows with indices 0 and 2 are indeed zeroed out
    assert_allclose(
        x,
        torch.tensor(
            [
                [0, 0, 0, 0],
                [4, 5, 6, 7],
                [0, 0, 0, 0],
                [12, 13, 14, 15],
            ],
            dtype=torch.float32,
        ),
        atol=tolerance,
    )


def test_reset_tensors_with_containers():
    tolerance = 1e-4

    # Generate tensors a, b, c, and d
    a = torch.tensor(
        [
            [0, 1],
            [2, 3],
            [4, 5],
        ],
        dtype=torch.float32,
    )

    b = torch.tensor(
        [
            [0, 10, 20],
            [30, 40, 50],
            [60, 70, 80],
        ],
        dtype=torch.float32,
    )

    c = torch.tensor(
        [
            [100],
            [200],
            [300],
        ],
        dtype=torch.float32,
    )

    d = torch.tensor([-1, -2, -3], dtype=torch.float32)

    # Put these tensors into a container
    my_tensors = [a, {"1": b, "2": (c, d)}]

    # Use `reset_tensors(...)` to clear the regions with indices 1 and 2 of each tensor in the container
    reset_tensors(my_tensors, [1, 2])

    # Assert that the reset operation was successful for each tensor in the container
    assert_allclose(
        a,
        torch.tensor(
            [
                [0, 1],
                [0, 0],
                [0, 0],
            ],
            dtype=torch.float32,
        ),
        atol=tolerance,
    )

    assert_allclose(
        b,
        torch.tensor(
            [
                [0, 10, 20],
                [0, 0, 0],
                [0, 0, 0],
            ],
            dtype=torch.float32,
        ),
        atol=tolerance,
    )

    assert_allclose(
        c,
        torch.tensor(
            [
                [100],
                [0],
                [0],
            ],
            dtype=torch.float32,
        ),
        atol=tolerance,
    )

    assert_allclose(d, torch.tensor([-1, 0, 0], dtype=torch.float32), atol=tolerance)


@torch.no_grad()
def test_policy_simple():
    tolerance = 1e-4

    # Make a torch module
    net = nn.Linear(5, 8)

    # Take its parameters
    parameters = parameter_vector(net)
    parameter_length = len(parameters)

    # Instantiate a policy from a clone of this torch module, and fill it with the same parameters
    policy = Policy(clone(net))
    policy.set_parameters(parameters)
    assert policy.parameter_length == parameter_length

    # Generate a random input
    x = torch.randn(5)

    # Pass the input through the original torch module and through the policy
    y1 = net(x)
    y2 = policy(x)

    # Assert that the output is the same
    assert_allclose(y2, y1, atol=tolerance)


@torch.no_grad()
def test_policy_batched():
    tolerance = 1e-4

    # Make a new torch module
    net = nn.Linear(5, 8)

    # Take its parameters
    parameters = parameter_vector(net)
    parameter_length = len(parameters)

    # Make a new Policy from a clone of the torch module, and fill it with the same parameters
    policy = Policy(clone(net))
    policy.set_parameters(parameters)
    assert policy.parameter_length == parameter_length

    # Declare a batch size, generate a batch of network parameters, and a batch of inputs
    batch_size = 10
    batch_of_parameters = torch.randn(batch_size, parameter_length)
    batch_of_x = torch.randn(batch_size, 5)

    # For each row in the parameter batch, fill the torch module's parameters with that row, pass the corresponding
    # input through the network, get its output
    y1 = []
    for i_row, row_of_x in enumerate(batch_of_x):
        row_of_parameters = batch_of_parameters[i_row]
        fill_parameters(net, row_of_parameters)
        y1.append(net(row_of_x))

    # Stack all the outputs generated by the torch module
    y1 = torch.stack(y1)

    # Fill the policy with the batch of parameters, and pass the batch of inputs through the policy
    policy.set_parameters(batch_of_parameters)
    y2 = policy(batch_of_x)

    # Assert that the policy's output matches the collected outputs of the torch module
    assert_allclose(y2, y1, atol=tolerance)


@torch.no_grad()
def test_policy_recurrent():
    tolerance = 1e-4

    net = LSTM(5, 8)

    parameters = parameter_vector(net)
    parameter_length = len(parameters)

    policy = Policy(clone(net))
    policy.set_parameters(parameters)
    assert policy.parameter_length == parameter_length

    x = torch.randn(3, 5)

    h1 = None
    y1 = None
    y2 = None
    for single_x in x:
        y1, h1 = net(single_x, h1)
        y2 = policy(single_x)

    assert_allclose(y2, y1, atol=tolerance)
    assert_allclose(policy.h[0], h1[0], atol=tolerance)
    assert_allclose(policy.h[1], h1[1], atol=tolerance)


@torch.no_grad()
def test_policy_recurrent_batched():
    tolerance = 1e-4
    batch_size = 10

    nets = [RNN(5, 8) for _ in range(batch_size)]

    policy = Policy(clone(nets[0]))

    parameters = torch.stack([parameter_vector(net) for net in nets])

    policy.set_parameters(parameters)
    assert policy.parameter_length == parameters.shape[1]

    num_steps = 20
    x = torch.randn(num_steps, batch_size, 5)
    dones_per_t = {
        5: [3, 7, 8],
        9: [True, True, False, True, False, False, True, True, False, False],
        14: [0, 1],
    }

    h1 = torch.zeros(batch_size, 8)
    y1 = torch.zeros(batch_size, 8)
    y2 = None
    for t in range(num_steps):
        single_x = x[t]

        for i_row, net in enumerate(nets):
            y1[i_row], h1[i_row] = net(single_x[i_row], h1[i_row])

        y2 = policy(single_x)

        if t in dones_per_t:
            mask = dones_per_t[t]
            h1[mask] = 0
            policy.reset(mask)

    assert_allclose(y2, y1, atol=tolerance)
    assert_allclose(policy.h, h1, atol=tolerance)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import torch
from functorch import make_functional_with_buffers
from torch import nn


def _shape_length(shape: tuple) -> int:
    """
    Return the number of elements implied by a shape tuple.

    Args:
        shape: A tuple or a torch.Size instance whose implied number of
            elements is being queried.
    Returns:
        An integer which represents the number of elements implied by the
        given shape tuple.
    """
    result = 1
    for item in shape:
        result *= int(item)
    return result


class ModuleExpectingFlatParameters:
    """
    A wrapper which brings a functional interface around a torch module.

    For obtaining the functional interface, this class internally uses
    the `functorch` library.

    Similar to `functorch.FunctionalModule`, `ModuleExpectingFlatParameters`
    turns a `torch.nn.Module` instance to a function which expects a new
    leftmost argument representing the parameters of the network.
    Unlike `functorch.FunctionalModule`, a `ModuleExpectingFlatParameters`
    instance, as its name suggests, expects the network parameters to be
    given as a 1-dimensional (i.e. flattened) tensor.
    Also, unlike `functorch.FunctionalModule`, an instance of
    `ModuleExpectingFlatParameters` is NOT an instance of `torch.nn.Module`.

    PyTorch modules with buffers can be wrapped by this class, but it is
    assumed that those buffers are constant. If the wrapped module changes
    the value(s) of its buffer(s) during its forward passes, most probably
    things will NOT work right.

    As an example, let us consider the following linear layer.

    ```python
    import torch
    from torch import nn

    net = nn.Linear(3, 8)
    ```

    The functional counterpart of `net` can be obtained via:

    ```python
    from evotorch.neuroevolution.net import ModuleExpectingFlatParameters

    fnet = ModuleExpectingFlatParameters(net)
    ```

    Now, `fnet` is a callable object which expects network parameters
    and network inputs. Let us call `fnet` with randomly generated network
    parameters and with a randomly generated input tensor.

    ```python
    param_length = fnet.parameter_length
    random_parameters = torch.randn(param_length)
    random_input = torch.randn(3)

    result = fnet(random_parameters, random_input)
    ```
    """

    def __init__(self, module: nn.Module, disable_autograd_tracking: bool = False):
        # Declare the variables which will store information regarding the parameters of the module.
        self.__param_shapes = []
        self.__param_length = 0
        self.__param_slices = []
        self.__num_params = 0
        self.__buffers = []

        # Iterate over the parameters of the module and fill the related information.
        i = 0
        j = 0
        for p in module.parameters():
            shape = p.shape
            self.__param_shapes.append(shape)

            length = _shape_length(shape)
            self.__param_length += length

            j = i + length
            self.__param_slices.append(slice(i, j))
            i = j

            self.__num_params += 1

        self.__fmodel, _, self.__buffers = make_functional_with_buffers(
            module, disable_autograd_tracking=bool(disable_autograd_tracking)
        )

        self.__buffers = list(self.__buffers)

    def __transfer_buffers(self, x: torch.Tensor):
        """
        Transfer the buffer tensors to the device of the given tensor.

        Args:
            x: The tensor whose device will also store the buffer tensors.
        """
        n = len(self.__buffers)
        for i in range(n):
            self.__buffers[i] = torch.as_tensor(self.__buffers[i], device=x.device)

    @property
    def buffers(self) -> tuple:
        """Get the stored buffers"""
        return self.__buffers

    @property
    def parameter_length(self) -> int:
        return self.__param_length

    def __call__(self, parameter_vector: torch.Tensor, x: torch.Tensor, h: Any = None) -> Any:
        """
        Call the wrapped module's forward pass procedure.

        Args:
            parameter_vector: A 1-dimensional tensor which represents the
                parameters of the tensor.
            x: The inputs.
            h: Hidden state(s), in case this is a recurrent network.
        Returns:
            The result of the forward pass.
        """
        if parameter_vector.ndim != 1:
            raise ValueError(
                f"Expected the parameters as 1 dimensional,"
                f" but the received parameter vector has {parameter_vector.ndim} dimensions"
            )
        if len(parameter_vector) != self.__param_length:
            raise ValueError(
                f"Expected a parameter vector of length {self.__param_length},"
                f" but the received parameter vector's length is {len(parameter_vector)}."
            )
        state_args = [] if h is None else [h]

        params = []
        for i in range(self.__num_params):
            param_slice = self.__param_slices[i]
            param_shape = self.__param_shapes[i]
            param = parameter_vector[param_slice].reshape(param_shape)
            params.append(param)

        # Make sure that the tensors are in the same device with x
        self.__transfer_buffers(x)

        # Run the functional module and return the results
        return self.__fmodel(params, self.__buffers, x, *state_args)


def make_functional_module(net: nn.Module) -> ModuleExpectingFlatParameters:
    """
    Wrap a torch module so that it has a functional interface.

    For obtaining a functional interface, this function internally uses the
    `functorch` library.

    Similar to `functorch.make_functional(...)`, this function turns a
    `torch.nn.Module` instance to a function which expects a new leftmost
    argument representing the parameters of the network.
    Unlike with `functorch.make_functional(...)`, the parameters of the
    network are expected in a 1-dimensional (i.e. flattened) tensor.

    PyTorch modules with buffers can be wrapped by this class, but it is
    assumed that those buffers are constant. If the wrapped module changes
    the value(s) of its buffer(s) during its forward passes, most probably
    things will NOT work right.

    As an example, let us consider the following linear layer.

    ```python
    import torch
    from torch import nn

    net = nn.Linear(3, 8)
    ```

    The functional counterpart of `net` can be obtained via:

    ```python
    from evotorch.neuroevolution.net import make_functional_module

    fnet = make_functional_module(net)
    ```

    Now, `fnet` is a callable object which expects network parameters
    and network inputs. Let us call `fnet` with randomly generated network
    parameters and with a randomly generated input tensor.

    ```python
    param_length = fnet.parameter_length
    random_parameters = torch.randn(param_length)
    random_input = torch.randn(3)

    result = fnet(random_parameters, random_input)
    ```

    Args:
        net: The `torch.nn.Module` instance to be wrapped by a functional
            interface.
    Returns:
        The functional wrapper, as an instance of
        `evotorch.neuroevolution.net.ModuleExpectingFlatParameters`.
    """
    return ModuleExpectingFlatParameters(net)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pickle
from argparse import ArgumentParser
from ast import literal_eval
from pathlib import Path
from time import sleep
from typing import Optional, Union

import gym
import numpy as np
import torch
from packaging.version import Version

from evotorch.neuroevolution.net.rl import reset_env, take_step_in_env

new_render_api = Version(gym.__version__) >= Version("0.26")


def make_env_for_rendering(*args, **kwargs):
    """
    Initialize a new gym environment with human-mode rendering.

    Beginning with gym 0.26, it is required to specify the rendering mode
    while initializing the environment. If the gym version is newer than
    or equal to 0.26, this function passes the keyword argument
    `render_mode="human"` to `gym.make(...)`.

    Args:
        args: Expected in the form of positional arguments. These are
            passed directly to `gym.make(...)`.
        kwargs: Expected in the form of keyword arguments. These are
            passed directly to `gym.make(...)`.
    Returns:
        The newly made gym environment.
    """
    if new_render_api:
        env_config = {"render_mode": "human"}
    else:
        env_config = {}

    env_config.update(kwargs)
    return gym.make(*args, **env_config)


def make_env_for_recording(*args, **kwargs):
    """
    Initialize a new gym environment with human-mode rendering.

    Beginning with gym 0.26, it is required to specify the rendering mode
    while initializing the environment. If the gym version is newer than
    or equal to 0.26, this function passes the keyword argument
    `render_mode="rgb_array"` to `gym.make(...)`.

    Args:
        args: Expected in the form of positional arguments. These are
            passed directly to `gym.make(...)`.
        kwargs: Expected in the form of keyword arguments. These are
            passed directly to `gym.make(...)`.
    Returns:
        The newly made gym environment.
    """
    if new_render_api:
        env_config = {"render_mode": "rgb_array"}
    else:
        env_config = {}

    env_config.update(kwargs)
    return gym.make(*args, **env_config)


def rgb_array_from_env(env: gym.Env) -> np.ndarray:
    """
    Render the current state of the environment into numpy array.

    Returns:
        The newly made numpy array containing the rendering result.
    """
    if new_render_api:
        return env.render()
    else:
        return env.render(mode="rgb_array")


def str_if_non_empty(s: Optional[str]) -> Optional[str]:
    if (s is None) or (isinstance(s, str) and (s == "")):
        return None
    else:
        return str(s)


def float_if_positive(x: Optional[float]) -> Optional[float]:
    if (x is None) or (x <= 0):
        return None
    else:
        return float(x)


def int_if_positive(x: Optional[int]) -> Optional[int]:
    if (x is None) or (x <= 0):
        return None
    else:
        return int(x)


def dict_if_non_empty(d: Optional[Union[str, dict]]) -> Optional[dict]:
    if d is None:
        return None
    elif isinstance(d, str):
        if d == "":
            return None
        else:
            return dict(literal_eval(d))
    elif isinstance(d, dict):
        return d
    else:
        raise TypeError(f"Object of unexpected type: {d}")


def main(
    fname: Union[str, Path],
    *,
    num_repeats: int = 1,
    t: Optional[float] = None,
    record_prefix: Optional[Union[str, Path]] = None,
    record_period: int = 1,
    extract: Optional[str] = None,
    set_in_env: Optional[str] = None,
    config: Optional[str] = None,
    env_name: Optional[str] = None,
):
    num_repeats = int(num_repeats)
    t = float_if_positive(t)
    record_prefix = str_if_non_empty(record_prefix)
    record_period = int(record_period)
    extract = str_if_non_empty(extract)
    set_in_env = dict_if_non_empty(set_in_env)
    config = dict_if_non_empty(config)
    env_name = str_if_non_empty(env_name)

    with open(fname, "rb") as f:
        loaded = pickle.load(f)

    if env_name is None:
        env_name = loaded["env_name"]
    policy = loaded["policy"]
    kwargs = {}
    if ("BulletEnv" in env_name) and (record_prefix is None) and (not new_render_api):
        kwargs["render"] = True
    if config is not None:
        kwargs.update(config)

    if record_prefix is None:
        env = make_env_for_rendering(env_name, **kwargs)
    else:
        env = make_env_for_recording(env_name, **kwargs)

    if set_in_env is not None:
        for k, v in set_in_env.items():
            setattr(env.unwrapped, k, v)

    def use_policy(observation):
        with torch.no_grad():
            action = policy(torch.as_tensor(observation, dtype=torch.float32)).numpy()
        if isinstance(env.action_space, gym.spaces.Discrete):
            action = np.argmax(action)
        return action

    if record_prefix is None:

        def render():
            env.render()
            if t is not None:
                sleep(t)

    else:
        scene_index = 0
        save_index = 0

        def render():
            nonlocal scene_index, save_index
            from matplotlib import pyplot as plt

            if scene_index % record_period == 0:
                img = rgb_array_from_env(env)
                if extract is not None:
                    extract_parts = extract.split(",")
                    x1 = int(extract_parts[0])
                    y1 = int(extract_parts[1])
                    x2 = int(extract_parts[2])
                    y2 = int(extract_parts[3])
                    img = img[y1:y2, x1:x2]
                scene_fname = "%s_%05d.png" % (record_prefix, save_index)
                plt.imsave(scene_fname, img)
                print(scene_fname)
                save_index += 1

            scene_index += 1

    repeat_iter = 0
    while repeat_iter < num_repeats or num_repeats <= 0:
        cumulative_reward = 0.0
        observation = reset_env(env)
        render()

        while True:
            action = use_policy(observation)
            observation, reward, done, info = take_step_in_env(env, action)
            render()
            cumulative_reward += float(reward)
            if done:
                break

        print(f"Repeat {repeat_iter}: {cumulative_reward}")
        repeat_iter += 1


if __name__ == "__main__":
    parser = ArgumentParser(
        description=(
            "This is a command-line tool for running, visualizing, and optionally recording the saved agents."
            " This tool works with the `.pickle` files saved by the training script named `rl.py`."
        )
    )

    parser.add_argument("fname", type=str, help="Name of the pickle file which stores the agent saved by rl.py")

    parser.add_argument("-n", type=int, default=1, help="Number of episodes over which the agent will be tested")

    parser.add_argument(
        "-t",
        type=float,
        default=-1.0,
        help=(
            "If given as a positive value, this much time, in seconds, will be introduced between each scene."
            " This option will be ignored if a non-empty string is given via the -r argument"
            " (a non-empty string for -r meaning that the recording mode is enabled)"
        ),
    )
    parser.add_argument(
        "-r",
        type=str,
        default="",
        help=(
            "When given as a non-empty string, this script will behave in recording mode."
            " With the recording mode enabled, each scene will be saved as a png file"
            " (instead of being shown to the screen)."
            " The prefix of the name of each created png file is determined by the string given for this argument."
        ),
    )

    parser.add_argument(
        "-p",
        type=int,
        default=1,
        help=(
            "When recording (i.e. when the argument -r is present), this argument determines the period"
            " for saving the scenes."
            " When given as a positive integer N, one scene from every N scenes will be saved."
            " By default, this is 1, which means there will be no skipped scenes."
        ),
    )

    parser.add_argument(
        "--extract",
        type=str,
        default="",
        help=(
            "When provided, this is expected as a string in the form 'x1,y1,x2,y2' where x and y values are integers."
            " In recording mode, this region will be extracted and this extracted region will be saved,"
            " instead of the whole scene."
            " For example, when the argument --extract '10,20,-30,-40' is given, the extracted region"
            " will be bounded by 10 from the left, 20 from the top, 30 from the right, and 40 from the bottom."
        ),
    )

    parser.add_argument(
        "--set",
        type=str,
        default="",
        help=(
            "When provided, this is expected as a Python dictionary."
            " Each item in this Python dictionary will be set as an attribute of the gym environment."
            " For example, for a PyBullet environment (e.g. bullet_envs:HumanoidBulletEnv-v0),"
            ' one might want to give the argument --set \'{"_render_width": 640, "_render_height": 480}\''
            " to set the attributes _render_width and _render_height as 640 and 480 respectively,"
            " therefore configuring the render size to 640x480 while recording."
        ),
    )

    parser.add_argument(
        "--config",
        type=str,
        default="",
        help=(
            "When provided, this is expected as a Python dictionary."
            " Each item in this Python dictionary will be sent to the environment as a keyword argument"
            " during its initialization phase."
        ),
    )

    parser.add_argument(
        "--env-name",
        type=str,
        default="",
        help=(
            "Name of the environment in which the policy will be tested."
            " If not specified, the environment name stored in the loaded pickle file will be used."
        ),
    )

    parsed = parser.parse_args()

    main(
        fname=parsed.fname,
        num_repeats=parsed.n,
        t=parsed.t,
        record_prefix=parsed.r,
        record_period=parsed.p,
        extract=parsed.extract,
        set_in_env=parsed.set,
        config=parsed.config,
        env_name=parsed.env_name,
    )
